import duckdb
from contextlib import contextmanager
from pathlib import PurePosixPath
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.fs import FilePath
from typing import Generator
from urllib.parse import urlparse


SUPPORTED_SCHEMES = set(['', 'file', 'http', 'https', 's3'])


class Column(BaseMetadataModel):
    type: str
    nullable: bool

class Columns(BaseMetadataModel):
    __root__: dict[str, Column] = {}


class FileColumnAnalyzer(AbstractAnalyzer):
    def __init__(self, url: str):
        # DuckDB doesn't understand 'file://' prefix, so remove it.
        self.url = url.removeprefix('file://')
        self.db = duckdb.connect()

    def analyze(
        self,
        path: FilePath,
    ) -> Columns | None:
        path_posix = PurePosixPath(str(path))
        url_and_path = self.url + str(path_posix)
        match path_posix.suffix:
            case ('.csv' | '.tsv'):
                self.db.execute(
                    "DESCRIBE SELECT * FROM read_csv_auto(?)",
                    [url_and_path],
                )
            case '.parquet':
                self.db.execute(
                    "DESCRIBE SELECT * FROM read_parquet(?)",
                    [url_and_path],
                )
            case _:
                return None
        columns_dict = {}
        for column_tuple in self.db.fetchall():
            name = column_tuple[0]
            type_ = column_tuple[1]
            nullable = column_tuple[2] == 'YES'
            columns_dict[name] = Column(type=type_, nullable=nullable)
        return Columns.parse_obj(columns_dict)


@contextmanager
def create_analyzer(
    url: str,
    **_,
) -> Generator[FileColumnAnalyzer, None, None]:
    scheme = urlparse(url).scheme
    if scheme in SUPPORTED_SCHEMES:
        yield FileColumnAnalyzer(url)
    else:
        raise ValueError(f"Unsupported url={url}")



import logging
import math
import pandas
from contextlib import contextmanager
from datetime import datetime
from json import dumps
from pathlib import PurePosixPath
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.db import TablePath, ViewPath
from recap.browsers.fs import FilePath
from typing import Any, Generator


log = logging.getLogger(__name__)


class ColumnProfile(BaseMetadataModel):
    count: int
    unique: int | None = None
    top: str | None = None
    freq: int | None = None
    mean: float | str | None = None
    std: float | str | None = None
    min: float | str | None = None
    p25: float | str | None = None
    p50: float | str | None = None
    p75: float | str | None = None
    p95: float | str | None = None
    p99: float | str | None = None
    p999: float | str | None = None
    max: float | str | None = None


class Profile(BaseMetadataModel):
    __root__: dict[str, ColumnProfile] = {}


class ProfileAnalyzer(AbstractAnalyzer):
    def __init__(
        self,
        url: str,
    ):
        self.url = url

    def analyze(
        self,
        path: TablePath | ViewPath | FilePath,
    ) -> Profile | None:
        path_posix = PurePosixPath(str(path))
        url_and_path = self.url + str(path_posix)
        df = pandas.DataFrame()
        match (path, path_posix.suffix):
            case (FilePath(), '.csv'):
                df = pandas.read_csv(url_and_path)
            case (FilePath(), '.tsv'):
                df = pandas.read_csv(url_and_path, sep='\t')
            case (FilePath(), '.json' | '.ndjson' | '.jsonl'):
                df = pandas.read_json(url_and_path, lines=True)
            case (FilePath(), '.parquet'):
                df = pandas.read_parquet(url_and_path)
            case (TablePath() | ViewPath(), _):
                # Meh.. try a SQL connection, I guess.
                # Types are pretty busted with structured matching. :(
                name = path.table if isinstance(path, TablePath) else path.view # pyright: ignore [reportGeneralTypeIssues]
                # TODO should sample the data here
                df = pandas.read_sql_table(
                    table_name=name,
                    schema=path.schema_, # pyright: ignore [reportGeneralTypeIssues]
                    con=self.url,
                )
        return self.analyze_dataframe(df) if not df.empty else None

    def analyze_dataframe(self, df: pandas.DataFrame) -> Profile:
        profile_dict = {}
        df_description = df.describe(
            percentiles=[.25, .5, .75, .95, .99, .999],
            include='all',
            # Get rid of FutureWarning
            datetime_is_numeric=True,
        )
        # Get rid of 'nan' for JSON.
        df_description = df_description.replace({math.nan: None})
        for name, stats_dict in df_description.to_dict().items():
            # Replace percentiles if they're set.
            formatted_percentiles = {
                'p25': stats_dict.pop('25%'),
                'p50': stats_dict.pop('50%'),
                'p75': stats_dict.pop('75%'),
                'p95': stats_dict.pop('95%'),
                'p99': stats_dict.pop('99%'),
                'p999': stats_dict.pop('99.9%'),
            } if '25%' in stats_dict else {}
            column_profile_dict: dict[str, Any] = (
                formatted_percentiles
                | stats_dict
            )
            # Replace datetime with ISO 8601 for JSON.
            column_profile_dict = dict([
                (k, v.isoformat())
                if isinstance(v, datetime)
                else (k, v)
                for k, v in column_profile_dict.items()
            ])
            # Pandas `top` can be a list or dict if input was JSON.
            # Convert it to a JSON string.
            if (
                column_profile_dict['top'] is not None
                and not isinstance(column_profile_dict['top'], str)
            ):
                column_profile_dict['top'] = dumps(column_profile_dict['top'])
            # Finally! Create the ColumnProfile.
            profile_dict[name] = ColumnProfile(**column_profile_dict)
        return Profile.parse_obj(profile_dict)


@contextmanager
def create_analyzer(url: str, **_) -> Generator['ProfileAnalyzer', None, None]:
    # TODO if URL is DB, create an engine that uses engine.* configs like db.py
    yield ProfileAnalyzer(url)



import logging
from .abstract import AbstractBrowser
from contextlib import contextmanager
from fsspec import AbstractFileSystem, get_fs_token_paths
from pathlib import PurePosixPath
from pydantic import Field
from recap.paths import CatalogPath
from typing import Any, Generator, Union
from urllib.parse import urlparse


log = logging.getLogger(__name__)


class FilesystemRootPath(CatalogPath):
    scheme: str
    name_: str = Field(alias='name')
    template = '/filesystems/{scheme}/instances/{name}'


class DirectoryPath(CatalogPath):
    # Path attr does not contain leading '/'. This is handled in the template.
    path: str
    template = '/{path:path}'


class FilePath(CatalogPath):
    # Path attr does not contain leading '/'. This is handled in the template.
    path: str
    template = '/{path:path}'


FilesystemBrowserPath = Union[
    DirectoryPath,
    FilePath,
]


class FilesystemBrowser(AbstractBrowser):
    """
    A browser that lists filesystem objects. FilesystemBrowser uses fsspec and
    its supported implementations
    (https://filesystem-spec.readthedocs.io/en/latest/api.html#built-in-implementations).

    FilesystemBrowser mirrors the directory structure in the filesystem.
    """

    def __init__(
        self,
        fs: AbstractFileSystem,
        base_path: str,
        root_: FilesystemRootPath,
    ):
        self.fs = fs
        self.base_path = base_path
        self.root_ = root_

    def children(self, path: str) -> list[FilesystemBrowserPath] | None:
        absolute_path = self.base_path + path
        paths = []
        if not self.fs.exists(absolute_path):
            return None
        if self.fs.isdir(absolute_path):
            # Force detail=True because gcsfs doesn't honor defaults.
            children = self.fs.ls(absolute_path, detail=True)
            for child in children:
                # Trim any duplicate // in the path.
                child_path = str(PurePosixPath('/', child['name']))
                # Remove base_path prefix since paths are relative.
                child_path = child_path.removeprefix(self.base_path)
                # Make child path relative for *Path models.
                child_path = child_path.lstrip('/')
                path_type = DirectoryPath if child['type'] == 'directory' \
                    else FilePath
                paths.append(path_type(path=child_path))
        return paths

    def root(self) -> FilesystemRootPath:
        return self.root_

    @staticmethod
    def default_root(url: str) -> FilesystemRootPath:
        parsed_url = urlparse(url)
        # Given `github://user:pass@main/test-data/test.csv`, return `github`.
        scheme = parsed_url.scheme.split('+')[0]
        # Given `github://user:pass@main/test-data/test.csv`, return `main`.
        name = parsed_url.netloc.split('@')[-1]
        return FilesystemRootPath(
            scheme=scheme or 'file',
            name=name or 'localhost',
        )


@contextmanager
def create_browser(
    url: str,
    name: str | None = None,
    storage_options: dict[str, Any] = {},
    **_,
) -> Generator[FilesystemBrowser, None, None]:
    """
    :param url: The URL to use for the filesystem. If the URL contains a path,
        the FilesystemBrowser will treat all paths relative to the URL path.
    :param name: The name to use in the FilesystemRootPath. If unspecified, the
        URL host is used (or 'localhost' for 'file' schemes).
    :param storage_options: Storage options **kwargs to pass on to the fsspec
        filesystem constructor.
    """

    default_root = FilesystemBrowser.default_root(url)
    fs, _, paths = get_fs_token_paths(url, storage_options=storage_options)

    assert len(paths) == 1, \
        f"Expected to get exactly 1 path from URL, but got paths={paths}"

    # Don't use DirFileSystem because it doesn't work properly with gcsfs.
    yield FilesystemBrowser(
        fs=fs,
        base_path=paths[0],
        root_=FilesystemRootPath(
            scheme=default_root.scheme,
            name=name or default_root.name_,
        ),
    )



import logging
from contextlib import contextmanager, ExitStack
from recap.analyzers import create_analyzer
from recap.analyzers.abstract import AbstractAnalyzer
from recap.browsers import create_browser as create_wrapped_browser
from recap.browsers.abstract import AbstractBrowser
from recap.paths import create_catalog_path, CatalogPath
from recap.plugins import load_analyzer_plugins, load_browser_plugins
from recap.typing import BrowserInspector
from typing import Any, Generator


log = logging.getLogger(__name__)


class AnalyzingBrowser(AbstractBrowser):
    """
    A special browser that ties an AbstractBrowser together with its compatible
    AbstractAnalyzers. Browsers and analyzers are deliberately unaware of one
    another, but something has to tie them together. That's this class.

    AnalyzingBrowser takes in a browser and its compatible analyzers. It
    provides all of the same functionality as the wrapped browser, but it also
    exposes an `analyze()` method, which runs all analyzers on a path and
    returns the analyzed metadata.

    This browser is deliberately not added to Recap's browser plugin
    entrypoint. It's more of a utility class for the CLI and crawler.
    """

    def __init__(
        self,
        browser: AbstractBrowser,
        analyzers: list[AbstractAnalyzer],
    ):
        self.browser = browser
        self.analyzers = analyzers
        self.root_str = str(browser.root())
        self.root_len = len(self.root_str)
        self.child_types = BrowserInspector(type(browser)).children_types()

    def children(self, path: str) -> list[CatalogPath] | None:
        return self.browser.children(path)

    def root(self) -> CatalogPath:
        return self.browser.root()

    def analyze(self, path: str) -> dict[str, Any] | None:
        if catalog_path := create_catalog_path(
            path,
            *self.child_types,
        ):
            return self._get_metadata(catalog_path)
        return None

    def _get_metadata(
        self,
        path: CatalogPath,
    ) -> dict[str, Any]:
        results = {}
        for analyzer in self.analyzers:
            log.debug(
                'Analyzing path=%s analyzer=%s',
                path,
                analyzer.__class__.__name__,
            )
            try: # EAFP
                if metadata := analyzer.analyze(path):
                    metadata_dict = metadata.dict(
                        by_alias=True,
                        exclude_none=True,
                        exclude_unset=True,
                        exclude_defaults=True,
                    )
                    # Have to unpack __root__ if it exists, sigh.
                    # https://github.com/pydantic/pydantic/issues/1193
                    metadata_dict = metadata_dict.get(
                        '__root__',
                        metadata_dict
                    )
                    results |= {metadata.key(): metadata_dict}
            except Exception as e:
                log.debug(
                    'Unable to process path with analyzer path=%s analyzer=%s',
                    path,
                    analyzer.__class__.__name__,
                    exc_info=e,
                )
        return results


@contextmanager
def create_browser(**config) -> Generator['AnalyzingBrowser', None, None]:
    """
    Create an AnalyzingBrowser that wraps an AbstractBrowser and
    AbstractAnalyzers that are compatible with the browser's child
    CatalogPaths.

    Note that the `plugin` config is currently ignored. This method simply
    searches for the first AbstractBrowser that doesn't throw an exception when
    its `create_browser` method is called.

    :param config: A **kwargs config for the browser to wrap.
    """

    analyzer_plugins = load_analyzer_plugins()
    browser_plugins = load_browser_plugins()

    with ExitStack() as stack:
        url = config.get('url')
        excludes = config.get('excludes', [])
        browser = None
        analyzers = []

        # Find a real AbstractBrowser to wrap
        for browser_name in browser_plugins.keys():
            try:
                browser_context_manager = create_wrapped_browser(
                    plugin=browser_name,
                    **config,
                )
                browser = stack.enter_context(browser_context_manager)

                # If we got this far, we found a browser. Stop looking.
                break
            except Exception as e:
                    log.debug(
                        'Skipped browser for url=%s name=%s',
                        url,
                        browser_name,
                        exc_info=e,
                    )

        assert browser, f"Found no browser for url={url}"

        # Find analyzers compatible with the real AbstractBrowser's config.
        for analyzer_name in analyzer_plugins.keys():
            if (analyzer_name not in excludes):
                try:
                    analyzer_context_manager = create_analyzer(
                        plugin=analyzer_name,
                        **config,
                    )
                    analyzer = stack.enter_context(
                        analyzer_context_manager,
                    )
                    analyzers.append(analyzer)
                except Exception as e:
                    log.debug(
                        'Skipped analyzer for url=%s name=%s',
                        url,
                        analyzer_name,
                        exc_info=e,
                    )

        if not analyzers:
            log.warn(
                "Found no metadata analyzers for url=%s. "
                "Crawling will only create directories.",
                url,
            )

        yield AnalyzingBrowser(browser, analyzers)



from contextlib import contextmanager
from frictionless import describe, Resource
from pathlib import PurePosixPath
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.fs import FilePath
from typing import Generator
from urllib.parse import urlparse


SUPPORTED_SCHEMES = set(['', 'file', 'http', 'https', 's3'])


class Column(BaseMetadataModel):
    type: str

class Columns(BaseMetadataModel):
    __root__: dict[str, Column] = {}


class FileColumnAnalyzer(AbstractAnalyzer):
    def __init__(self, url: str):
        self.url = url

    def analyze(
        self,
        path: FilePath,
    ) -> Columns | None:
        path_posix = PurePosixPath(str(path))
        url_and_path = self.url + str(path_posix)
        match path_posix.suffix:
            case ('.csv' | '.tsv' | '.parquet'):
                resource = describe(url_and_path)
                if isinstance(resource, Resource):
                    columns_dict = {}
                    for field in resource.schema.fields:
                        columns_dict[field.name] = Column(type=field.type)
                    return Columns.parse_obj(columns_dict)
            case ('.json' | '.ndjson' | '.jsonl'):
                resource = describe(path=url_and_path, format='ndjson')
                if isinstance(resource, Resource):
                    columns_dict = {}
                    for field in resource.schema.fields:
                        columns_dict[field.name] = Column(type=field.type)
                    return Columns.parse_obj(columns_dict)
            case _:
                return None


@contextmanager
def create_analyzer(
    url: str,
    **_,
) -> Generator[FileColumnAnalyzer, None, None]:
    scheme = urlparse(url).scheme
    if scheme in SUPPORTED_SCHEMES:
        if scheme == '':
            # Frictionless is paranoid about absolute paths. Use a file scheme
            # so that it allows them.
            url = f"file://{url}"
        yield FileColumnAnalyzer(url)
    else:
        raise ValueError(f"Unsupported url={url}")



from contextlib import contextmanager
from fsspec import AbstractFileSystem
from genson import SchemaBuilder
from json import loads
from pathlib import PurePosixPath
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.fs import create_browser, FilePath
from typing import Any, Generator


class JsonSchema(BaseMetadataModel):
    __root__: dict[str, Any]


class FileColumnAnalyzer(AbstractAnalyzer):
    def __init__(
        self,
        fs: AbstractFileSystem,
        base_path: str,
        sample: int | None = 1024,
    ):
        """
        :param fs: An fsspec filesystem to read files from.
        :param base_path: The `path` portion of the URL to read from. Acts as
            the root path for child paths.
        :param sample: If set, read the first N rows of a JSON file when
            inferring its schema. Sometimes preferable for performance reasons.
        """

        self.fs = fs
        self.base_path = base_path
        self.sample = sample

    def analyze(
        self,
        path: FilePath,
    ) -> JsonSchema | None:
        builder = SchemaBuilder()
        absolute_path_posix = PurePosixPath(self.base_path, str(path).lstrip('/'))
        if (
            absolute_path_posix.suffix == '.json'
            or absolute_path_posix.suffix == '.ndjson'
            or absolute_path_posix.suffix == '.jsonl'
        ):
            with self.fs.open(str(absolute_path_posix), 'rt') as f:
                line_count = 0
                while (
                    (obj := f.readline().strip())
                    and (
                        not self.sample
                        or line_count < self.sample
                    )
                ):
                    builder.add_object(loads(obj))
                    line_count += 1
            return JsonSchema.parse_obj(builder.to_schema())
        return None


@contextmanager
def create_analyzer(
    url: str,
    sample: int | None = 1024,
    **config,
) -> Generator[FileColumnAnalyzer, None, None]:
    # TODO This is hacky. Shoulld factor out FS and patch creation.
    with create_browser(url=url, **config) as browser:
        yield FileColumnAnalyzer(browser.fs, browser.base_path, sample)



import fnmatch
import logging
from contextlib import contextmanager
from pathlib import PurePosixPath
from recap.browsers.analyzing import AnalyzingBrowser, create_browser
from recap.catalogs.abstract import AbstractCatalog
from recap.paths import CatalogPath, RootPath
from typing import Any, Generator


log = logging.getLogger(__name__)


class Crawler:
    """
    Crawler does three things:

    1. Browses the configured infrastructure
    2. Analyzes the infrastructure's data to generate metadata
    3. Stores the metadata in a catalog
    """

    def __init__(
        self,
        browser: AnalyzingBrowser,
        catalog: AbstractCatalog,
        recursive: bool = True,
        filters: list[str] = [],
    ):
        """
        :param browser: AnalyzingBrowser to use for listing children and
            analyzing metadata.
        :param catalog: The catalog where the crawler will create directories
            and store metadata.
        :param recursive: Whether the crawler should recurse into
            subdirectories when crawling.
        :param filters: Path filter to include only certain paths. Recap uses
            Unix filename pattern matching as defined in Python's fnmatch
            module. Filtered paths are relative to the browser (excluding the
            browser's root).
        """

        self.browser = browser
        self.catalog = catalog
        self.recursive = recursive
        self.filters = filters
        self.exploded_filters = self._explode_filters(filters)

    def crawl(self):
        log.info('Beginning crawl root=%s', self.browser.root())
        path_stack: list[CatalogPath] = [RootPath()]

        while len(path_stack) > 0:
            relative_path = str(path_stack.pop())
            full_path_posix = PurePosixPath(
                str(self.browser.root()),
                relative_path[1:],
            )

            log.info("Crawling path=%s", relative_path)
            self.catalog.touch(str(full_path_posix))

            # 1. Read and save metadata for path if filters match.
            if (
                self._matches(relative_path, self.filters)
                and (metadata := self.browser.analyze(relative_path))
            ):
                self._write_metadata(full_path_posix, metadata)

            # 2. Add children (that match filter) to path_stack.
            children = self.browser.children(relative_path) or []
            filtered_children = filter(
                lambda p: self._matches(str(p), self.exploded_filters),
                children,
            )
            if self.recursive:
                path_stack.extend(filtered_children)

            # 3. Remove deleted children from catalog.
            self._remove_deleted(full_path_posix, children)

        log.info('Finished crawl root=%s', self.browser.root())

    def _matches(
        self,
        relative_path: str,
        filters: list[str],
    ) -> bool:
        """
        Check if a path matches any filters.

        :returns: True if path matches a filter or if filters is empty.
        """

        for filter in filters:
            if fnmatch.fnmatch(relative_path, filter):
                return True
        return False if filters else True

    def _write_metadata(
        self,
        full_path_posix: PurePosixPath,
        metadata: dict[str, Any],
    ):
        """
        Write a metadata dictionary to a path in the catalog.
        """

        log.debug(
            'Writing metadata path=%s metadata=%s',
            full_path_posix,
            metadata,
        )
        self.catalog.write(str(full_path_posix), metadata, True)

    def _remove_deleted(
        self,
        full_path_posix: PurePosixPath,
        instance_children: list[CatalogPath],
    ):
        """
        Compares the path's children in the browser vs. what is currently in
        the catalog. Deletes all children that appear in the catalog, but no
        longer appear in the browser. This behavior removes children that used
        to exist in data infrastructure, but have been deleted since the last
        crawl.
        """

        catalog_children = self.catalog.ls(str(full_path_posix)) or []
        instance_children_names = [c.name() for c in instance_children]
        # Find catalog children that are not in the browser's children.
        deleted_children = [
            catalog_child
            for catalog_child in catalog_children
            if catalog_child not in instance_children_names
        ]
        for child in deleted_children:
            path_to_remove = str(PurePosixPath(full_path_posix, child))
            log.debug('Removing deleted path from catalog: %s', path_to_remove)
            self.catalog.rm(path_to_remove)

    def _explode_filters(self, filters: list[str]) -> list[str]:
        """
        Returns a list of paths that bread-crumb from the filter all the way
        back to root. For example:

            filters=[
                '/**/schemas/my_db/tables/foo*'
            ]
            returns=[
                '/**',
                '/**/schemas',
                '/**/schemas/my_db',
                '/**/schemas/my_db/tables',
                '/**/schemas/my_db/tables/foo*',
            ]

        We need to do this so that parents match the filter and crawling
        reaches the wild-carded children.
        """

        exploded_filters = []
        for filter in filters:
            fragments = filter.split('/')
            partial_path = PurePosixPath('/')
            for fragment in fragments:
                partial_path = PurePosixPath(partial_path, fragment)
                exploded_filters.append(str(partial_path))
        return exploded_filters


@contextmanager
def create_crawler(
    url: str,
    catalog: AbstractCatalog,
    **config,
) -> Generator['Crawler', None, None]:
    with create_browser(url=url, **config) as browser:
        yield Crawler(browser, catalog, **config)
